/*  
    Copyright (C) 2004 Julia Handl
    Email: Julia.Handl@gmx.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


#include "som.h"
#include "random.h"

extern long idum;

int * perm;

int max(int x, int y) {
    if (x > y) return x;
    else return y;
}

int min(int x, int y) {
    if (x < y) return x;
    else return y;
}



/* Constructor */
som::som(conf * par, databin<USED_DATA_TYPE> * bin, evaluation * e):
    clalg(par, bin, e) {
  
 

  // allocate weights
    weight = new USED_DATA_TYPE ** [par->imax_som];
    for (int i=0; i<par->imax_som; i++) {
	weight[i] = new USED_DATA_TYPE * [par->jmax_som];
	for (int j=0; j<par->jmax_som; j++) {
	    weight[i][j] = new USED_DATA_TYPE[par->bindim];
	}
    }
}

/* Destructor */
som::~som() {
    // free all allocated memory
    for (int i=0; i<par->imax_som; i++) {
	for (int j=0; j<par->jmax_som; j++) {
	    delete [] weight[i][j];
	}
	delete[] weight[i];
    }
    delete weight;
    
  
}


/* Intialise weight vectors (uniformly random)
   within the region "covered" by the data */
void som::init() {
  
    // initialise weight vectors
    for (int i=0; i<par->imax_som; i++) {
	for (int j=0; j<par->jmax_som; j++) {
	    for (int k=0; k<par->bindim; k++) {
		weight[i][j][k] = bin->minvalue[k]+ran0(&idum)*(bin->maxvalue[k]-bin->minvalue[k]);
	    }
	}
    }
}

/* 1D SOM, after training: determine the output neuron for 
   each data item and generate clustering object */
void som::classify() {

    // determine output neuron for each input item
   clust = new clustering(par);
   clust->init(par->kclusters);
   data<USED_DATA_TYPE> * datacentre = new data<USED_DATA_TYPE>[par->kclusters](par);
   int * ctr = new int[par->kclusters];
   for (int i=0; i<par->kclusters; i++) {
       ctr[i] = 0;

   }

  
 
    for (int k=0; k<par->binsize; k++) {

	data<USED_DATA_TYPE> & input = (*bin)[k];

	
	double mindist = 1e100;
	int iptr = 0;
	int jptr = 0;

	for (int i=0; i<par->imax_som; i++) {
	    for (int j=0; j<par->jmax_som; j++) {

		data<USED_DATA_TYPE> * temp = new data<USED_DATA_TYPE>(par, (weight[i][j]));
		double dist = input.distanceto(*temp);
		delete temp;

 		if (dist < mindist) {
		    mindist = dist;
		    iptr = i;
		    jptr = j;
		}
	    }
	}
	datacentre[iptr*par->imax_som+jptr].add((*bin)[k]);
	(*clust)[k] = jptr;
	ctr[jptr]++;
    }
    for (int k=0; k<par->kclusters; k++) {
 	datacentre[k].div(ctr[k]);
	clust->newcentre(k, &(datacentre[k]));
    }

}



/* Training phase */
void som::run() {

    double ns = par->ns0_som;
    double lr = par->lr0_som;


    USED_DATA_TYPE mindist;
    int iptr;
    int jptr;

    perm = new int[par->binsize];

    for (int generation_ctr=1; generation_ctr <= par->gen_som; generation_ctr++) {

      permutation(); 
	for (int t=0; t<par->binsize; t++) {
	    data<USED_DATA_TYPE> & stimulus = (*bin)[perm[t]];
	    
	    mindist = 1e100;
	    iptr = 0;
	    jptr = 0;
	    
	    // compute Euclidean distances and determine
	    // winner neuron
	    for (int i=0; i<par->imax_som; i++) {
		for (int j=0; j<par->jmax_som; j++) {
		    
		    data<USED_DATA_TYPE> * temp = new data<USED_DATA_TYPE>(par, (weight[i][j]));
		    double dist = stimulus.distanceto(*temp);
	    
		    delete temp;
		    
		    if (dist < mindist) {
		      
			mindist = dist;
			iptr = i;
			jptr = j;
		    }
		}
	    }
	    
	    // adapt weights
    
	    for (int i = max(0,int(iptr-ns)); i <= min(int(iptr+ns),par->imax_som-1); i++) {
		for (int j = max(0,int(jptr-ns)); j <= min(int(jptr+ns),par->jmax_som-1); j++) {
		  int ii = i;
		  int jj = j;
		    double z = square(ii-iptr) + square(jj-jptr);		    
		    double factor = lr * exp(-z / ns);
		    for (int k=0; k<par->bindim; k++) {
			double diff = stimulus[k] - weight[ii][jj][k];
			weight[ii][jj][k] += factor*diff;
		    }
		}
	    }
	    
	    // adapt lr and ns
	    lr = par->lr0_som * (1.0-double((generation_ctr-1)*par->binsize+t)/double(par->gen_som*par->binsize));
	    ns = par->ns1_som + (par->ns0_som - par->ns1_som)* (1.0-double((generation_ctr-1)*par->binsize+t)/double(par->gen_som*par->binsize));
	    
	}

    }
    par->ns0_som = ns;

}


double som::square(double x) {
    return x*x;
}



// generate a permutation
void som::permutation() {

    for (int i=0; i<par->binsize; i++) {
	perm[i] = i;
    }
    for (int i=0; i<par->binsize; i++) {
	int j = int(ran0(&idum)*(par->binsize));
	int k = int(ran0(&idum)*(par->binsize));

	int temp = perm[i];
	perm[i] = perm[k];
	perm[k] = temp;
    }
}
