
/*  
    Copyright (C) 2004 Julia Handl
    Email: Julia.Handl@gmx.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/



#include "sota.h"
#include "random.h"
#include "conf.h"

extern long idum;

// learning rates
double lr_w = 0.01;
double lr_a = 0.005;
double lr_s = 0.001;


double sota::square(double x) {
    return x*x;
}

double sota::myabs(double x) {
  if (x < 0) return -x;
  else return x;
}




/* Destructor for an individual node */
node::~node(){
  delete weight;
  delete member;
}


/* Constructor for an individual node */
node::node(conf * par) {
    weight = new float[par->bindim];
    for (int i=0; i<par->bindim;i++) {
      weight[i] = 0.0;
    }
    member = new int[par->binsize];
    var = 0.0;
    left = NULL;
    right = NULL;
    anc = NULL;
    d = 0;
    ctr = 0;
  }


/* Constructor */
sota::sota(conf * par, databin<USED_DATA_TYPE> * bin, evaluation * e):
    clalg(par, bin, e) {
  
  root = new node(par);
  

}

/* Recursive deletion of nodes */
void sota::del(node * root) {
  if (root->left != NULL) 
  del(root->left);
  if (root->right != NULL) 
  del(root->right);
  delete root;
  
}


/* Destructor */
sota::~sota() {
  if (root->left != NULL) 
    del(root->left);
  if (root->right != NULL) 
    del(root->right);
  delete root;
}



/* Recursive computation of variance */
void sota::setdiv(conf * par, node * current) {

  // calculate diversity
  if (current->left != NULL) 
    setdiv(par, current->left);
  if (current->right != NULL)
    setdiv(par, current->right);
  if (current->left == NULL && current->right == NULL) {
    current->var = (current->ctr > 0)? current->d/double(current->ctr) : 0;
    }
}


/* Intialisation of the root node to average data vector */
void sota::init() {
  
  for (int i=0; i<par->binsize; i++) {
    for (int j=0; j<par->bindim; j++) {
      root->weight[j] += (*bin)[i][j];
    }
    root->member[i] = 1;
  }
  for (int j=0; j<par->bindim; j++) {
    root->weight[j] /= double(par->binsize);
  }

  for (int i=0; i<par->binsize; i++) {
    double d = 0;
    for (int j=0; j<par->bindim; j++) {
      d += square((*bin)[i][j] - root->weight[j]);
    }
    root->d += sqrt(d);
  }
  root->var /= root->d/double(par->binsize);
  root->id = 0;
  root->ctr = par->binsize;

}



/* Recursive determination of cluster id */
void sota::calculate(conf * par, node * go, double * epsilon, int * id) {
  if (go->right != NULL)
    calculate(par, go->left, epsilon, id);
  if (go->right != NULL)
    calculate(par, go->right, epsilon, id);

  if (go->left == NULL && go->right == NULL) {
    go->id = *id;
    (*id)++;
  }

}

/* Find node with largest intra cluster variance */
node * sota::maxvar(conf * par,node * xx, double * var) {
    node * maxj;
    node * next = xx;
    double maxvarx = -1e10;
    double maxvary = -1e10;
    double maxv = -1e10;
    node * x = NULL;
    node * y = NULL;
    if (next->left != NULL) 
      x = maxvar(par, next->left, &maxvarx);
    if (next->right != NULL)
      y = maxvar(par, next->right, &maxvary);

    if (maxvarx >= maxvary) {
      maxj = x;
      maxv = maxvarx;
    }
    else {
      maxj = y;
      maxv = maxvary;
    }

    if (next->left == NULL && next->right == NULL) {
      maxj = next;
      maxv = next->var;
     
    }
    *var = maxv;
    return maxj;
}

/* Find node closest to input vector */
node * sota::mindist(conf * par, node * xx, double * var, data<USED_DATA_TYPE> * stimulus, int t) {
  node * minnode;
  node * current = xx;
  double mindx = 1e10;
  double mindy = 1e10;
  double mind = 1e10;
  node * x = NULL;
  node * y = NULL;
  if (current->left != NULL) 
    x = mindist(par, current->left, &mindx, stimulus, t);
  if (current->right != NULL)
    y = mindist(par, current->right, &mindy, stimulus, t);
  
  if (mindx < mindy) {
    minnode = x;
    mind = mindx;
  }
  else {
    minnode = y;
    mind = mindy;
  }
  
  if (current->left == NULL && current->right == NULL ) {
    minnode = current;
    data<USED_DATA_TYPE> * temp = new data<USED_DATA_TYPE>(par, current->weight);
    mind = stimulus->distanceto(*temp);
    delete temp;
  }

  *var = mind;
  return minnode;
}

/* Determine membership */
node * sota::cluster(conf * par, node * xx, int t) {
  node * minnode = NULL;
  node * x = NULL;
  node * y = NULL;

  if (xx->left != NULL) 
    x = cluster(par, xx->left,t);
  if (xx->right != NULL)
    y = cluster(par, xx->right, t);
  
  
  if (xx->left == NULL && xx->right == NULL && xx->member[t] == 1) {
    minnode = xx;
  }

  if (minnode == NULL && x != NULL) {
    minnode = x;
  }
  if (minnode == NULL && y != NULL) {
    minnode = y;
  }

 
  return minnode;
}




/* Main routine of SOTA algorithm */
void sota::run() {


  // Increase number of clusters by one in each cycle
  for (int i=1; i<par->kclusters; i++) {
    double epsilon=-1;
    double epsilon_old=-2;


    // Find node to split (node with largest resource value)
    node * maxj;
    node * next = root;
  
    double maxvarx = -1e10;
    double maxvary = -1e10;
    double maxv = -1e10;
    node * x = NULL;
    node * y = NULL;
    if (next->left != NULL) 
      x = maxvar(par, next->left, &maxvarx);
    if (next->right != NULL)
      y = maxvar(par, next->right, &maxvary);

    if (maxvarx > maxvary) {
      maxj = x;
      maxv = maxvarx;
    }
    else {
      maxj = y;
      maxv = maxvary;
    }

    if (next->left == NULL && next->right == NULL) {
      maxj = next;
      maxv = next->var;
     }

     // Create and initialize the two children 
    maxj->left = new node(par);
    maxj->right = new node(par);
    maxj->left->anc = maxj;
    maxj->right->anc = maxj;


    for (int l=0;l<par->bindim;l++) {
      maxj->left->weight[l] = maxj->weight[l];
      maxj->right->weight[l] = maxj->weight[l];
    }
    for (int l=0;l<par->binsize;l++) {
      maxj->left->member[l] = 0;
      maxj->right->member[l] = 0;
    }

    int member[par->binsize];
    bool interrupt = false;
 
    // Training of the two sister nodes
    for (int generation_ctr=1; generation_ctr <= 1000; generation_ctr++) {
    
      if (myabs((epsilon-epsilon_old)/epsilon_old) < 0.0001 ) {
	break;
      }

      maxj->left->var = 0.0;
      maxj->right->var = 0.0;
      maxj->left->d = 0.0;
      maxj->right->d = 0.0;
      maxj->left->ctr = 0;
      maxj->right->ctr = 0;

      for (int t=0; t<par->binsize; t++) {
	  member[t] = maxj->member[t];

	  if (maxj->member[t] == 0){
	    continue;
	  }
	  data<USED_DATA_TYPE> * stimulus = &((*bin)[t]);
	
	  double mind = 1e10;
	  node * minnode;
	

	  data<USED_DATA_TYPE> * temp1 = new data<USED_DATA_TYPE>(par, maxj->left->weight);
	  data<USED_DATA_TYPE> * temp2 = new data<USED_DATA_TYPE>(par, maxj->right->weight);
	  double mind1 = stimulus->distanceto(*temp1);
	  double mind2 = stimulus->distanceto(*temp2);
	  delete temp1;
	  delete temp2;


	  if (mind1 <= mind2) {
	    minnode = maxj->left;
	    mind = mind1;
	  }
	  else {
	    minnode = maxj->right;
	    mind = mind2;
	  }

	  // Remember error
	  minnode->d += mind;
	  minnode->ctr = minnode->ctr+1;

	
	  // Adapt weights
	  for (int l=0; l<par->bindim; l++) {
	    minnode->weight[l] = minnode->weight[l] + lr_w * ((*bin)[t][l] - minnode->weight[l] );
	  }
	  node * sister = NULL;
	  if (minnode->anc->left == minnode) {
	    sister = minnode->anc->right;
	  }
	  else {
	    sister = minnode->anc->left;
	  }
	  
	  if (sister->left ==NULL && sister->right ==NULL) {
    
	    for (int l=0; l<par->bindim; l++) {
	      sister->weight[l] = sister->weight[l] + lr_s * ((*bin)[t][l] - sister->weight[l] );
	    }
	    for (int l=0; l<par->bindim; l++) {
	      minnode->anc->weight[l] = minnode->anc->weight[l] + lr_a * ( (*bin)[t][l] - minnode->anc->weight[l]);
	    }
	  } 
      }

      // Compute relative error
      epsilon_old = epsilon;
      int id = 0;
      if (root->left != NULL)
	calculate(par, root->left, &epsilon, &id);
      if (root->right != NULL)
	calculate(par, root->right, &epsilon, &id);
      if (root->left == NULL && root->right == NULL) {
	root->id = id;
	id++;
      }

      maxj->left->var = (maxj->left->ctr > 0)? maxj->left->d/double(maxj->left->ctr) : 0;
      maxj->right->var = (maxj->right->ctr > 0)? maxj->right->d/double(maxj->right->ctr) : 0;
      epsilon = maxj->left->var+maxj->right->var;
    }



    // Postprocessing: assign data items of the split parent node to other clusters
    maxj->left->d = maxj->left->var = 0.0;
    maxj->right->d = maxj->right->var = 0.0;
    maxj->right->ctr = 0;
    maxj->left->ctr = 0;


    // Find the right level
    node * currentCell = maxj;
    int levelctr = 0;
    while (currentCell->anc != NULL && levelctr < 5) {
      currentCell = currentCell->anc;
      levelctr++;
    }

    node * current = currentCell;


    // Assign new memberships to all members of the split parent node
    for (int t=0; t<par->binsize; t++) {
      if (member[t] == 0){
	continue;
      }

      data<USED_DATA_TYPE> * stimulus = &((*bin)[t]);
      
      node * minnode;
      node * current = currentCell;
      double mindx = 1e100;
      double mindy = 1e100;
      double mind = 1e100;
      node * x = NULL;
      node * y = NULL;
      if (current->left != NULL) 
	x = mindist(par, current->left, &mindx, stimulus, t);
      if (current->right != NULL)
	y = mindist(par, current->right, &mindy, stimulus, t);
     
      if (mindx < mindy) {
	minnode = x;
	mind = mindx;
      }
      else {
	minnode = y;
	mind = mindy;
      }
     
      if (current->left == NULL && current->right == NULL) {
	minnode = current;
	data<USED_DATA_TYPE> * temp = new data<USED_DATA_TYPE>(par, current->weight);
	mind = stimulus->distanceto(*temp);
	delete temp;
      }

      minnode->d += mind;
      minnode->ctr += 1;
      minnode->member[t] = 1;
    }

    // calculate intra-cluster variance (resource value) for all nodes
    if (current->left != NULL) 
      setdiv(par, current->left);
    if (current->right != NULL)
      setdiv(par, current->right);
    if (current->left == NULL && current->right == NULL) {
      current->var = (current->ctr > 0)? current->d/double(current->ctr) : 0;
    }

  }
}


/* SOTA, after training: read the cluster assignment for all data items
   and generate clustering object */
void sota::classify() {


   clust = new clustering(par);
   clust->init(par->kclusters);
   data<USED_DATA_TYPE> * datacentre = new data<USED_DATA_TYPE>[par->kclusters](par);
   int * ctr = new int[par->kclusters];
   for (int i=0; i<par->kclusters; i++) {
       ctr[i] = 0;

   }

   // read the cluster assignment for each input item   
   for (int t=0; t<par->binsize; t++) {

     data<USED_DATA_TYPE> * stimulus = &((*bin)[t]);
     
     node * x = NULL;
     node * y = NULL;
     node * minnode = NULL;
     node * current = root;

     if (current->left != NULL) 
       x = cluster(par, current->left, t);
     if (current->right != NULL)
       y = cluster(par, current->right, t);
        
     if (current->left == NULL && current->right == NULL && current->member[t] == 1) {
       minnode = current;
     }

     if (minnode == NULL && x != NULL) {
       minnode = x;
     }
     if (minnode == NULL && y != NULL) {
       minnode = y;
     }
     
     datacentre[minnode->id].add((*bin)[t]);
     (*clust)[t] = minnode->id;
     ctr[minnode->id]++;
   }
  
   for (int k=0; k<par->kclusters; k++) {
     cerr << ctr[k] << " ";
     if (ctr[k] > 0) datacentre[k].div(ctr[k]);
     clust->newcentre(k, &(datacentre[k]));
   }
   cerr << endl;

}	


	  



